@@ -182,6 +182,15 @@ g_io_channel_read (GIOChannel *channel,
   g_return_val_if_fail (channel != NULL, G_IO_ERROR_UNKNOWN);
   g_return_val_if_fail (bytes_read != NULL, G_IO_ERROR_UNKNOWN);
 
+  if (count == 0)
+    {
+      if (bytes_read)
+        *bytes_read = 0;
+      return G_IO_STATUS_NORMAL;
+    }
+
+  g_return_val_if_fail (buf != NULL, G_IO_ERROR_UNKNOWN);
+
   status = channel->funcs->io_read (channel, buf, count, bytes_read, &err);
 
   error = g_io_error_get_from_g_error (status, err);
@@ -237,7 +246,7 @@ g_io_channel_write (GIOChannel  *channel,
  *        file).
  * 
  * Sets the current position in the #GIOChannel, similar to the standard library
- * function <function>fseek()</function>. 
+ * function fseek(). 
  * 
  * Return value: %G_IO_ERROR_NONE if the operation was successful.
  *
@@ -284,7 +293,7 @@ g_io_channel_seek  (GIOChannel   *channel,
  * g_io_channel_new_file:
  * @filename: A string containing the name of a file.
  * @mode: One of "r", "w", "a", "r+", "w+", "a+". These have
- *        the same meaning as in <function>fopen()</function>.
+ *        the same meaning as in fopen().
  * @error: A location to return an error of type %G_FILE_ERROR.
  *
  * Open a file @filename as a #GIOChannel using mode @mode. This
@@ -549,7 +558,7 @@ g_io_channel_error_from_errno (gint en)
 
 #ifdef EFAULT
     case EFAULT:
-      g_warning("File descriptor outside valid address space.\n");
+      g_warning("Buffer outside valid address space.\n");
       return G_IO_CHANNEL_ERROR_FAILED;
 #endif
 
@@ -739,7 +748,7 @@ g_io_channel_set_flags (GIOChannel *channel,
  * The values of the flags %G_IO_FLAG_IS_READABLE and %G_IO_FLAG_IS_WRITEABLE
  * are cached for internal use by the channel when it is created.
  * If they should change at some later point (e.g. partial shutdown
- * of a socket with the UNIX <function>shutdown()</function> function), the user
+ * of a socket with the UNIX shutdown() function), the user
  * should immediately call g_io_channel_get_flags () to update
  * the internal values of these flags.
  *
