@@ -44,6 +44,7 @@
 #endif
 
 #include "glib.h"
+#include "gprintf.h"
 #include "gprintfint.h"
 
 #ifdef G_OS_WIN32
@@ -51,7 +52,7 @@
 #endif
 
 /* do not include <unistd.h> in this place since it
- * inteferes with g_strsignal() on some OSes
+ * interferes with g_strsignal() on some OSes
  */
 
 static const guint16 ascii_table_data[256] = {
@@ -181,29 +182,13 @@ g_stpcpy (gchar       *dest,
 
 gchar*
 g_strdup_vprintf (const gchar *format,
-		  va_list      args1)
+		  va_list      args)
 {
-  gchar *buffer;
-#ifdef HAVE_VASPRINTF
-  if (_g_vasprintf (&buffer, format, args1) < 0)
-    buffer = NULL;
-  else if (!g_mem_is_system_malloc ()) 
-    {
-      gchar *buffer1 = g_strdup (buffer);
-      free (buffer);
-      buffer = buffer1;
-    }
-#else
-  va_list args2;
+  gchar *string = NULL;
 
-  G_VA_COPY (args2, args1);
+  g_vasprintf (&string, format, args);
 
-  buffer = g_new (gchar, g_printf_string_upper_bound (format, args1));
-
-  _g_vsprintf (buffer, format, args2);
-  va_end (args2);
-#endif
-  return buffer;
+  return string;
 }
 
 gchar*
@@ -229,7 +214,8 @@ g_strconcat (const gchar *string1, ...)
   gchar	  *concat;
   gchar   *ptr;
 
-  g_return_val_if_fail (string1 != NULL, NULL);
+  if (!string1)
+    return NULL;
 
   l = 1 + strlen (string1);
   va_start (args, string1);
@@ -448,7 +434,7 @@ g_ascii_strtod (const gchar *nptr,
 
       if (fail_pos)
 	{
-	  if (fail_pos > decimal_point_pos)
+	  if (fail_pos - copy > decimal_point_pos - nptr)
 	    fail_pos = (char *)nptr + (fail_pos - copy) - (decimal_point_len - 1);
 	  else
 	    fail_pos = (char *)nptr + (fail_pos - copy);
@@ -2185,6 +2171,13 @@ g_strsplit (const gchar *string,
   return str_array;
 }
 
+/**
+ * g_strfreev:
+ * @str_array: a %NULL-terminated array of strings to free.
+
+ * Frees a %NULL-terminated array of strings, and the array itself.
+ * If called on a %NULL value, g_strfreev() simply returns. 
+ **/
 void
 g_strfreev (gchar **str_array)
 {
@@ -2553,3 +2546,30 @@ g_str_has_prefix (const gchar  *str,
   return strncmp (str, prefix, prefix_len) == 0;
 }
 
+
+/**
+ * g_strip_context:
+ * @msgid: a string
+ * @msgval: another string
+ * 
+ * An auxiliary function for gettext() support (see Q_()).
+ * 
+ * Return value: @msgval, unless @msgval is identical to @msgid and contains
+ *   a '|' character, in which case a pointer to the substring of msgid after
+ *   the first '|' character is returned. 
+ *
+ * Since: 2.4
+ **/
+G_CONST_RETURN gchar *
+g_strip_context  (const gchar *msgid, 
+		  const gchar *msgval)
+{
+  if (msgval == msgid)
+    {
+      const char *c = strchr (msgid, '|');
+      if (c != NULL)
+	return c + 1;
+    }
+  
+  return msgval;
+}
