@@ -957,8 +957,22 @@ g_get_any_init (void)
 	  }
 	if (pw)
 	  {
-	    g_user_name = g_strdup (pw->pw_name);
-	    g_real_name = g_strdup (pw->pw_gecos);
+ 	    g_user_name = g_strdup (pw->pw_name);
+
+	    if (pw->pw_gecos) 
+	      {
+		gchar **gecos_fields;
+		gchar **name_parts;
+
+		/* split the gecos field and substitute '&' */
+		gecos_fields = g_strsplit (pw->pw_gecos, ",", 0);
+		name_parts = g_strsplit (gecos_fields[0], "&", 0);
+		pw->pw_name[0] = g_ascii_toupper (pw->pw_name[0]);
+		g_real_name = g_strjoinv (pw->pw_name, name_parts);
+		g_strfreev (gecos_fields);
+		g_strfreev (name_parts);
+	      }
+
 	    if (!g_home_dir)
 	      g_home_dir = g_strdup (pw->pw_dir);
 	  }
@@ -993,20 +1007,6 @@ g_get_any_init (void)
 	g_user_name = g_strdup ("somebody");
       if (!g_real_name)
 	g_real_name = g_strdup ("Unknown");
-      else
-	{
-	  gchar *p;
-
-	  for (p = g_real_name; *p; p++)
-	    if (*p == ',')
-	      {
-		*p = 0;
-		p = g_strdup (g_real_name);
-		g_free (g_real_name);
-		g_real_name = p;
-		break;
-	      }
-	}
     }
 }
 
